#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  player.py
#  
#  Copyright 2019 Stefan Born <born@math.tu-berlin.de>
#  
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#  
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#  
#  

import pygame
import random
import math

from random import shuffle

from constants import *



class Player:
    
    stamina = 0
    survived_steps = 0
    alive = True
    x = 0
    y = 0 
    world = None # verweist auf die world-map 
    
    def mutate(self, initial=False):
        '''Diese Methode hat zwei Varianten:
        Ist der Parameter initial==True,
        so wird eine Anfangskonfiguration erzeugt.
        Ist der Parameter initial==False,
        so wird ein gegebenes 'Genom' mutiert.
        '''
        raise NotImplementedError
        
    def set_start_position(self,x,y):
        
        raise NotImplementedError
        
    def move(self):
        '''Das ist die Handlung des Wesens. Gegenwärtig 
        kann es sich nur bewegen.'''
        
        raise NotImplementedError
        

class RandomPlayer(Player):

    #Initialisiert den Spieler
    def __init__(self,x,y,speed,name):
        self.x=x #x-Pos
        self.y=y #y-Pos
        self.r=0
        self.angle = 0 #Laufwinkel  
        self.status = True #Aktiv oder nicht
        self.chunkpos = [0,0] #Speichert die Position des Chunks
        
        self.viewpoint = 0
        self.sichtfeld = 0   #Prozentuale Anzahl von Objekten im Sichtfeld
        self.viewangle = 60  #Winkel des Sichtfelds
        self.v = False       #sind Objekte im Sichtfeld
        
        self.size = 1 #Größe
        self.sight = 10 #Sichtweite
        self.fat = 0    #Die Dicke der Fettschicht der Kreaturen
        self.speed = 1 #Geschwindigkeit
      
        self.inner_temprature = 0  #Die Innere Temperatur der Kreaturen
        self.skin_temprature = 0   #Die Hauttemperatur der Kreaturen
        
        self.rect=pygame.Rect(int(self.x)-int(self.size),int(self.y)-int(self.size),self.size*2,self.size*2)#Rechteck in Pygame        
        self.name=name #ID  
        self.alive = True    #Ist die Kreatur am Leben
        self.world = None

        self.poly_move=[]    #die Koeffizienten des Polynoms zur Bestimmung der Schrittreichweite (hat die Form [a,b] mit a*x**b)
        self.poly_angle=[]   #die Koeffizienten des Polynoms zur Bestimmung der Winkeländerung (genau wie bei Schrittreichweite)

        self.survived_steps = 0 #Anzahl der Überlebten Züge
        self.collected_food = 0    #Die Anzahl des gesammelten Futters
        self.stamina = START_STAMINA #Ausdauer
        self.generation = 0     #Generation der Kreatur
        
        self.memory = []       #Erinnerung der Kreatur 
        self.memory_length = 0 #größtmögliche Länge der Erinnerung
        

    def set_start_position(self,x,y):
        self.x=x #x-Pos
        self.y=y
        self.rect=pygame.Rect(int(self.x)-int(self.size),int(self.y)-int(self.size),self.size*2+self.size,self.size*2+self.size)#Rechteck in Pygame  
    
    #mutate mutiert die Eigenschaften und movement_parameters die Verhaltensparameter der Kreatur 
    #(hier Verhalten auf Basis von Polynomen)
    def mutate(self,initial=False):
        '''Diese Methode hat zwei Varianten:
        
        Ist der Parameter initial==True,
        so wird eine Anfangskonfiguration erzeugt.
        Ist der Parameter initial==False,
        so wird ein gegebenes 'Genom' mutiert.
        '''
        
        if initial:
            self.speed = random.randint(3,6)
            self.size  = random.randint(3,6)
            self.sight = random.randint(10,30)*self.size
            self.angle = random.randint(0,360)
            self.fat   = random.randint(0,max(int(self.size/2),1))
            self.memory_length = random.randint(0,5)
            
        else:
            self.speed = self.speed+random.choice(range(-9,10))*0.001*self.speed
            self.sight = self.sight+random.choice(range(-9,10))*0.01*self.sight
            self.size  = self.size+random.choice(range(-9,10))*0.01*self.size
            self.fat   = self.fat+random.choice(range(-9,10))*0.01
            self.memory_length = self.memory_length+random.choice([-1,0,1])
            if self.fat <0:
                self.fat==0
            elif self.fat >self.size/2:  #Fettschicht kann nicht dicker sein als ein viertel der Größe
                self.fat=self.fat/2
            
        self.stamina = START_STAMINA+self.fat*self.size
        self.max_stamina = self.stamina*STAMINA_KOEFFIZIENT
        self.death = random.randint(1,MAX_DEATH)
    
    def movement_parameters(self,initial=False):
        
        if initial:  #wählt eine Länge l des Polynoms und erzeugt dieses
            l=random.randint(2,10)
       #Die eingaben für die ersellung des Polynoms sind Länge, Minimale und Maximale Potenz, Minimale und Maximale Koeffizienten,
       #Minimale und Maximale Ausgabe des Polynoms sowie die aktuellen Länge des Gedächtnisses für evtl. tiefere Erinnerungspolynome
            self.poly_move=polynom(l,0,10,0,10,0,10,["m","s"],self.memory_length)
            l=random.randint(2,10)
            self.poly_angle=polynom(l,0,10,0,10,0,10,["m","s"],self.memory_length)
            l=random.randint(2,10)
            self.poly_input=polynom(l,0,10,0,10,0,10,["f","u","k"],0,True)
        else:  #verändert das Polynom in Länge und dem Wert der Koeffizienten
            self.poly_move.change_poly
            self.poly_angle.change_poly
            self.poly_input.change_poly
                
        self.complexity_movement = self.poly_move.length
        self.complexity_angle = self.poly_angle.length
        self.death=random.randint(2000,3000)

    def looking_forward(self,objliste):
        self.v = False
        self.angle=self.angle%360 
        viewpoint_x =self.x
        viewpoint_y =self.y
        
        def distance(x1,y1,x2,y2):
            return ((x1-x2)**2+(y1-y2)**2)**(1/2)
        for obj in objliste:    
            dist = distance(obj.x,obj.y,self.x,self.y)
            
            if dist > self.sight*(2)**1/2:
                self.v = False
            else:
                points = []
                points.append([obj.x-obj.size,obj.y-obj.size])
                points.append([obj.x+obj.size,obj.y-obj.size])
                points.append([obj.x-obj.size,obj.y+obj.size])
                points.append([obj.x+obj.size,obj.y+obj.size])
                i = 0
                
                angles=[]
                for point in points:
                    if(self.sight**2>((viewpoint_x-point[0])**2+(viewpoint_y-point[1]**2))):
                        if(viewpoint_x==point[0]):
                            if(viewpoint_y>point[1]):
                                angle=270
                            else:
                                angle=90
                        else:
                            angle=math.atan((viewpoint_y-point[1])/(viewpoint_x-point[0]))/math.pi*180
                            #print(angle)
                            if(viewpoint_x>point[0]):
                                angle+=180
                            elif(angle<0):
                                angle+=360
                        #print("c")
                        angles.append(angle)
                y=(-(-((self.sight**2)-(viewpoint_x-points[0][0])**2)**(1/2)-viewpoint_y))
                y2=(-(((self.sight**2)-(viewpoint_x-points[0][0])**2)**(1/2)-viewpoint_y))
                #print(y)
                #print(y2)
                if((self.sight**2>(viewpoint_x-points[0][0])**2)and points[0][1]+2*obj.size>y and(points[0][1]<y)):
                    if(viewpoint_x==points[0][0]):
                        if(viewpoint_y>y):
                            angle=270
                        else:
                            angle=90
                    else:
                        angle=math.atan((viewpoint_y-y)/(viewpoint_x-points[0][0]))/math.pi*180
                        if(viewpoint_x>points[0][0]):
                            angle+=180
                        elif(angle<0):
                            angle+=360
                    #print(y,"a",angle)
                    angles.append(angle)
                if(self.sight**2>(viewpoint_x-points[0][0])**2 and points[0][1]+2*obj.size>y2 and(points[0][1]<y2)):
                    if(viewpoint_x==points[0][0]):
                        if(viewpoint_y>y2):
                            angle=270
                        else:
                            angle=90
                    else:
                        angle=math.atan((viewpoint_y-y2)/(viewpoint_x-points[0][0]))/math.pi*180
                        if(viewpoint_x>points[0][0]):
                            angle+=180
                        elif(angle<0):
                            angle+=360
                    #print(y2,"a",angle)
                    angles.append(angle)
                
                y=(-(-((self.sight**2)-(viewpoint_x-points[1][0])**2)**(1/2)-viewpoint_y))
                y2=(-(((self.sight**2)-(viewpoint_x-points[1][0])**2)**(1/2)-viewpoint_y))
                #print(y)
                #print(y2)
                if((self.sight**2>(viewpoint_x-points[1][0])**2)and points[1][1]+2*obj.size>y and(points[1][1]<y)):
                    if(viewpoint_x==points[0][0]):
                        if(viewpoint_y>y):
                            angle=270
                        else:
                            angle=90
                    else:
                        angle=math.atan((viewpoint_y-y)/(viewpoint_x-points[1][0]))/math.pi*180
                        if(viewpoint_x>points[1][0]):
                            angle+=180
                        elif(angle<0):
                            angle+=360
                    #print(y,"b",angle)
                    angles.append(angle)
                if((self.sight**2>(viewpoint_x-points[1][0])**2)and points[0][1]+2*obj.size>y2 and(points[0][1]<y2)):
                    if(viewpoint_x==points[0][0]):
                        if(viewpoint_y>y2):
                            angle=270
                        else:
                            angle=90
                    else:
                        angle=math.atan((viewpoint_y-y)/(viewpoint_x-points[0][0]))/math.pi*180
                        if(viewpoint_x>points[0][0]):
                            angle+=180
                        elif(angle<0):
                            angle+=360
                    #print(y2,"b",angle)
                    angles.append(angle)
                x=(-(-(((self.sight**2)-(viewpoint_y-points[0][1])**2)**(1/2))-viewpoint_x))
                x2=(-(((self.sight**2)-(viewpoint_y-points[0][1])**2)**(1/2)-viewpoint_x))
                #print(x)
                #print(x2)
                if((self.sight**2>=(viewpoint_y-points[0][1])**2)and points[0][0]+(2*obj.size)>x and(points[0][0]<x)):
                    if(viewpoint_x==x):
                        angle=90
                    else:
                        angle=math.atan((viewpoint_y-points[0][1])/(viewpoint_x-x))/math.pi*180
                    if(viewpoint_x>x):
                        angle+=180
                    elif(angle<0):
                        angle+=360
                    #print(x,"c",angle)
                    angles.append(angle)
                if((self.sight**2>(viewpoint_y-points[0][1])**2)and points[0][0]+(2*obj.size)>x2 and(points[0][0]<x2)):
                    if(viewpoint_x==x2):
                        angle=90
                    else:
                        angle=math.atan((viewpoint_y-points[0][1])/(viewpoint_x-x2))/math.pi*180
                    if(viewpoint_x>x2):
                        angle+=180
                    elif(angle<0):
                        angle+=360
                    #print(x2,"c",angle)
                    angles.append(angle)
                x=(-(-((self.sight**2)-(viewpoint_y-points[2][1])**2)**(1/2)-viewpoint_x))
                x2=(-(((self.sight**2)-(viewpoint_y-points[2][1])**2)**(1/2)-viewpoint_x))
            
                if((self.sight**2>=(viewpoint_y-points[2][1])**2)and points[2][0]+(2*obj.size)>x and(points[2][0]<x)):
                    if(viewpoint_x==x):
                        if(viewpoint_y>points[2][1]):
                            angle=270
                        else:
                            angle=90
                    else:
                        angle=math.atan((viewpoint_y-points[2][1])/(viewpoint_x-x))/math.pi*180
                        if(viewpoint_x>x):
                            angle+=180
                        elif(angle<0):
                            angle+=360
                    #print(x,"d",angle)
                    angles.append(angle)
                if((self.sight**2>=(viewpoint_y-points[2][1])**2)and points[2][0]+2*obj.size>x2 and(points[2][0]<x2)):
                    if(viewpoint_x==x2):
                        if(viewpoint_y>points[2][1]):
                            angle=270
                        else:
                            angle=90
                    else:
                        angle=math.atan((viewpoint_y-points[2][1])/(viewpoint_x-x2))/math.pi*180
                        if(viewpoint_x>x):
                            angle+=180
                        elif(angle<0):
                            angle+=360
                    #print(x2,"d",angle)
                    angles.append(angle)

            if(len(angles)!=0):
                if(max(angles)-min(angles)>250):
                    for angle in angles:
                        if(angle>250):
                            angle=angle-360
                if((max(angles)>(self.angle+self.viewangle))):
                    if(min(angles)>(self.angle+self.viewangle)):
                        self.sichtfeld=0
                    elif(min(angles)<(self.angle-self.viewangle)):
                        self.sichtfeld=1
                    else:
                        self.sichtfeld=((self.angle+self.viewangle)-min(angles))/(self.viewangle*2)
                elif(min(angles)<(self.angle-self.viewangle)):
                    if(max(angles)<(self.angle-self.viewangle)):
                        self.sichtfeld=0                        
                    #print(self.angle-self.viewangle)
                    else:
                        self.sichtfeld=(max(angles)-(self.angle-self.viewangle))/(self.viewangle*2)
                else:
                    self.sichtfeld=((max(angles)-(min(angles)))/(self.viewangle*2))
            else:
                if((max(angles)-min(angles))>180):
                    for angle in angles:
                        if(angle>270):
                            angle=angle-360
               # print(points)
                self.sichtfeld=0
                #print(self.x,self.y)
                #print(angles)
                #self.sichtfeld=max(angles)-min(angles)
                #print(self.sichtfeld)       
            #return((max(angles)-min(angles)/viewangle)
            #print(max(angles)-min(angles)/(2*self.viewangle))
            #print(self.sichtfeld)



                
    def sightcheck(self,dist,alpha):
        self.objangle = alpha
        #print(dist)
        if dist<=self.sight:
            #print(self.angle%360,alpha)
            if self.angle-self.viewangle<0:
                if (self.angle%360-self.viewangle)+360 <= alpha or alpha <= self.angle+self.viewangle:
                    self.v = True
                    self.angle=alpha
                    print("2",self.v)
            elif self.angle-self.viewangle>360:
                if (self.angle%360-self.viewangle)-360 <= alpha or alpha <= self.angle+self.viewangle:
                    self.v = True
                    #print(self.v)
                    self.angle=alpha
            else:
                if self.angle%360-self.viewangle <= alpha <= self.angle+self.viewangle:
                    self.v = True
                    #print(self.v)
                    self.angle=alpha
        
        self.sichtfeld = 0.9
    
    
    
    #bewegt den Spieler
    def move(self):
        
        #Überprüfe, ob die Maximale Stamina überschritten wird und verringere die Stamina, sollte dem so sein
        if self.stamina>self.max_stamina:
            self.stamina=self.max_stamina
        
        r=0          #Größe des "Schrittes"
        new_angle=0  #Größe der Winkeländerung
        poly_move=self.poly_move
        poly_angle=self.poly_angle
        
        
        info = self.world.get_info(self) #Liste mit möglichen Objekten in Sichtweite
        
  #Sichtweite in der Darstellung verzehnfacht
        self.sight=self.sight*10

        
        if len(info)!=0:
            self.looking_forward(info)         #"Nach vorne schauen"
            futter_bedeckung=self.sichtfeld    #Wert der Futterbedeckung des sichtfeldes 
                
        else:
            futter_bedeckung=0
            
        w=0
        for m in self.poly_input.poly:
            if m[0]<0:
                a=random.randint(0,10)
            else:
                a=m[0]

            if m[1]<0:
                b=random.randint(0,10)
            else:
                b=m[1]
            
            if m[2]=="f":
                x=futter_bedeckung
            if m[2]=="u":
                x=MAPSIZE-self.x+MAPSIZE-self.y
            if m[2]=="k":
                x=futter_bedeckung
            
            w+=w+a*x**b
            
        while w>self.poly_input.max_output:
            w=self.poly_input.max_output-(w-self.poly_input.max_output)
        while w<self.poly_input.min_output:
            w=w+(self.poly_input.min_output-w)
            
            
        r_reaction=self.poly_move.reaction(w,self.memory)      #Reaktion der Kreatur bezüglich der schrittweite
        angle_reaction=self.poly_angle.reaction(w,self.memory) #Reaktion der Kreatur bezüglich der Winkeländerung
        
        if len(self.memory)>self.memory_length: #"vergessen" der ältesten Erinnerung wenn die maximale Gedächnisslänge erreicht wurde
            del (self.memory[-1])
        self.memory.append([r_reaction,angle_reaction,futter_bedeckung])  #hinzufügen der ebigen Reaktion zur Erinnerung
            
            
        r=self.speed*(r_reaction/10)              #Endgültige ermittlung der Schrittweite und
        self.angle=self.angle+36*(angle_reaction) #der Winkeländerung
            
        self.sight=self.sight/10

        if self.survived_steps>self.death:
            self.alive=False
        
        elif self.stamina >= self.speed+self.size**2+self.sight*0.2: #Die Bewegung erfolgt nur, wenn genügend Stamina vorhanden ist
            self.x+=math.sin(0.01745329*self.angle)*r #x und
            self.y+=math.cos(0.01745329*self.angle)*r #y änderung
            self.rect=pygame.Rect(int(self.x)-int(self.size),int(self.y)-int(self.size),self.size*2+self.size,self.size*2+self.size)
            
            ## Es ist wichtig, eine Modellierung zu machen, die nicht
            ## schon ein eindeutiges Evolutionsergebnis nahelegt.
            ## Schneller zu gehen, sollte mehr Energie kosten.
            ## Ein größeres Areal abzufressen sollte eine Energie 
            ## kosten, die proportional zum Areal ist.
            
            self.skin_temprature=self.inner_temprature-self.fat
            
            #Der Energieverlust durch Temperatur ergibt sich aus dem Umfang, also dem freiligenden Teil des Körpers
            energy_loss_through_temprature=(self.skin_temprature-TEMPRATURE)*8*self.size  
            
            if energy_loss_through_temprature<0:
                energy_loss_through_temprature=0
            
            #Der Stamina-verlust an sich:
            self.stamina -=  MIN_STAMINA_LOSS_PER_STEP+0.5*(abs(self.speed)**2+self.size)
            
            self.survived_steps += 1 
        else:
            self.alive = False

    
#Die Klasse polynom ist nur releant, wenn die Simulation auf Basis von Polynomen erfolgen soll   
# Mit dieser Klasse werden sowohl die normalen als auch die Erinnerungspolynome erstellt 
class polynom:
    length=0     #Startlänge des Polynoms
    poly=[]      #Das Polynom an sich
    min_potenz=0 #minimale Startpotenz
    max_potenz=0 #maximale Startpotenz
    min_koeffizient=0  #min Startkoeffizient
    max_koeffizient=0  #max Startkoeffizient
    min_output=0  #min Ausgabe
    max_output=0  #max Ausgabe
    
    def __init__(self,l,minp,maxp,mink,maxk,mino,maxo,dif_monos,memory_length,deep_poly=False):
        self.length=l
        self.min_potenz=minp
        self.max_potenz=maxp
        self.min_koeffizient=mink
        self.max_koeffizient=maxk
        self.min_output=mino
        self.max_output=maxo
        self.poly=[]
        self.deep_poly=deep_poly  #ist das Polynom ein tiefes Polynom, also ein Polynom zur einbindung der Erinnerung
        
        n=0
        while n<self.length:  #erstellen des Polynoms
            a=random.choice(range(self.min_koeffizient,self.max_koeffizient+1))
            b=random.choice(range(self.min_potenz,self.max_potenz+1))
            c=random.choice(dif_monos) #c legt die Art des Inputs fest mit hier mr=Erinnerung der Schrittweite, ma=Erinnerung der Winkeländerung und mf=Erinnerung der Futterabdeckung
            if self.deep_poly==False and c=="m":  #Für Erinnerungspolynome
                new_poly=polynom(memory_length,0,10,0,10,0,1,["mf","mr","ma"],0,True) #Erinnerungspolynom wird erstellt
                self.poly.append([a,b,c,new_poly])
            else:
                self.poly.append([a,b,c,[]])
            n+=1
    
    def change_poly(self,memory_length):  #Änderung, also Mutation des Polynoms
        n=0
        self.A=[]
        if self.length<2:
            self.length=random.choice(range(0,2))+self.length
        else:
            self.length=random.choice(range(-1,2))+self.length
        while self.length>len(self.poly):
            a=random.choice(range(self.min_koeffizient,self.max_koeffizient+1))
            b=random.choice(range(self.min_potenz,self.max_potenz+1))
            c=random.choice(dif_monos_list) #c legt die Art des Inputs fest mit hier mr=Erinnerung der Schrittweite, ma=Erinnerung der Winkeländerung und mf=Erinnerung der Futterabdeckung
            if self.deep_poly==False and c==1:  #Für Erinnerungspolynome
                new_poly=polynom(memory_length,0,10,0,10,0,1,3,0,True) #Erinnerungspolynom wird erstellt
                self.poly.append([a,b,c,new_poly])
            else:
                self.poly.append([a,b,c,[]])
        random.shuffle(self.poly)
        while n<self.length:
            mutation=random.choice(range(0,MUTATIONSWAHRSCHEINLICHKEIT_VERHALTEN))
            if mutation==0:
                a=random.choice(range(-5,6))
                b=random.choice(range(-9,10))
                aneu=(self.poly[n][0])+0.01*a
                bneu=(self.poly[n][1])+0.01*b
                self.A.append([aneu,bneu,self.poly[n][2],self.poly[n][3]])
            else:
                self.A.append(self.poly[n])

            n+=1  
        self.poly=self.A
        
    def deep_reaction(self,l):  #Reaktion für erinnerungspolynome
        if self.deep_poly:
            c=0
            N=min(len(self.poly),len(l))
            n=0
            
            while n<N:

                if self.poly[n][0]<0:
                    a=random.randint(0,10)
                else:
                    a=self.poly[n][0]

                if self.poly[n][1]<0:
                    b=random.randint(0,10)
                else:
                    b=self.poly[n][1]

                if self.poly[n][2]=="mr":
                    x=l[n][0]
                    
                elif self.poly[n][2]=="ma":
                    x=l[n][1]    
                
                elif self.poly[n][2]=="mf":
                    x=l[n][2]

                c=c+a*x**b
                n+=1
            #c=c/len(poly)
            while c>self.max_output:
                c=self.max_output-(c-self.max_output)
            while c<self.min_output:
                c=c+(self.min_output-c)

            return c    
                
        else:
            return 0
            
    
    def reaction(self,s,memory):  #Reaktion der Kreatur auf einen bestimmten Umstand für das normale Polynom
        c=0
        
        for el in self.poly:

            if el[0]<0:
                a=random.randint(0,10)
            else:
                a=el[0]

            if el[1]<0:
                b=random.randint(0,10)
            else:
                b=el[1]

            if el[2]=="s":
                x=s
            elif el[2]=="m":
                x=el[3].deep_reaction(memory)
                
            c=c+a*x**b
        #c=c/len(poly)
        while c>self.max_output:
            c=self.max_output-(c-self.max_output)
        while c<self.min_output:
            c=c+(self.min_output-c)

        return c    
    
    def show(self):  #zeigt das Polynom als ganzes, mit allen tieferen Polynomen
        self.A = []
        for p in self.poly:
            if p[2]=="s":
                self.A.append(p)
            if p[2]=="m":
                deep=p[3].poly
                self.A.append([p[0],p[1],p[2],deep])

        
        
def main(args):
    return 0

if __name__ == '__main__':
    import sys
    sys.exit(main(sys.argv))        
    
    